/***************************************************************/
/* Berechnung der Zahl pi = 3.1415926... auf Basis der Formel: */
/*							       */
/* pi	  4    1     1	4      1      1  4	1	       */
/* --- = (- - ---) - -(--- - -----) + -(--- - -----) - + ...   */
/*  4	  5   239    3 5^3   239^3    5 5^5   239^5	       */
/*							       */
/* von Reinhard Scharnagl, 4000 Duesseldorf, Feb.-April 1992,  */
/* unter der Benutzung von PureC, in der Version Feb 03 1992   */
/* (Tab = 8)						       */
/***************************************************************/

/* Anpassungen */

#ifdef __STDC__
#define _(parms) parms	 /* ANSI C version */
#include <stdlib.h>
#else
#define _(parms) ()	 /* old C version */
#define void /**/
#endif

#ifdef __TURBOC__
#pragma warn -use
#endif

/* eingebundene Help-Dateien */

#include <stdio.h>

/* lokale Prototypen */

void big_set _((unsigned *field, unsigned value,       int length));
int  big_div _((unsigned *field, unsigned divisor,     int length));
void big_add _((unsigned *field, unsigned *summand,    int length));
void big_mul _((unsigned *field, unsigned factor,      int length));
void big_out _((unsigned *field, int	  digits,      int length));
void big_cpy _((unsigned *field, unsigned *source,     int length));
void big_sub _((unsigned *field, unsigned *subtrahend, int length));

typedef unsigned BUFFER[6810*sizeof(unsigned)];

/***************************************************************/
/* Aufruf "CALC_PI.TTP nnn" mit 0 < nnn < 32768 Dezimalstellen */
/***************************************************************/

int main _((int ac, char **av));
int main(ac,av)
  int ac;
  char **av;
{ register unsigned divisor;
  register int huge;
  int digits, arg;
  unsigned *sum,		     /* langes Rechenfeld 1 */
	   *quo,		     /* langes Rechenfeld 2 */
	   *buf;		     /* langes Rechenfeld 3 */

  $TABLE GetRam 	 /* genuegend Speicher verfuegbar ? */
  /*--------------------------------------------------------*/
  $C ((sum = (unsigned *)malloc(sizeof(BUFFER))) == 0L)
				/* erster Puffer zu Gross ? */	$ Y N N N
     ((quo = (unsigned *)malloc(sizeof(BUFFER))) == 0L)
			       /* zweiter Puffer zu Gross ? */	$ . Y N N
     ((buf = (unsigned *)malloc(sizeof(BUFFER))) == 0L)
			       /* dritter Puffer zu Gross ? */	$ . . Y N
  $A printf("Speicher reicht nicht!\n\n");   /* Fehler-Text */	$ X X X .
     return 1;				/* Error, Abbruch ! */	$ X X X .
  $END
  
  $TABLE AskParm	  /* Nachfrage, wenn ohne Parameter */
  /*--------------------------------------------------------*/
  $C (ac < 2)			       /* keine Argumente ? */	$ Y N
  $A printf("gewuenschte Genauigkeit? ");    /* User fragen */	$ X .
     scanf("%s", av[ac++] = (char *)sum);   /* hole Antwort */	$ X .
     arg = 0;		     /* starte vor erstem Parameter */	$ X X
  $END

  $TABLE DoActArg	 /* verarbeite anstehende Argumente */
  /*--------------------------------------------------------*/
  $C (++arg >= ac)		    /* kein Argument mehr ? */	$ Y N
  $A return 0;			    /* OK, alles erledigt ! */	$ X .
  $END

  $TABLE ReadArg		/* lese Ziffern-Anforderung */
  /*--------------------------------------------------------*/
  $C (sscanf(av[arg],"%d",&digits))   /* Zahl einlesen OK ? */	$ Y Y N
     (digits <= 0)		    /* illegale Forderung ? */	$ Y N .
  $A printf("Eingabe fehlerhaft: %s\n\n", av[arg]);
					     /* Fehler-Text */	$ X . X
     goto DoActArg;		   /* Anforderung skippen ! */	$ X . X
  $END

  $TABLE StartCalc		    /* Start der Berechnung */
  /*--------------------------------------------------------*/
     printf("Berechnung von pi mit %d Stellen:\n\n",digits);
				     /* Titelzeile ausgeben */
     huge = (int)((digits*1000L)/(301L*8*sizeof(unsigned))+3);
		/* berechne obere Schranke der Rechenfelder */
     big_set(sum,     0, huge);     /* init. Summe auf	1.0 */
  $END

  $TABLE DoCalcPh1		  /* Phase 1 der Berechnung */
  /*--------------------------------------------------------*/
     big_set(quo, 4*4*5, huge);     /* init. Quot. auf 80.0 */
     for (divisor = 1; ; divisor += 2)	  /* Iter.schleife: */
  $C ( big_div(quo, 5*5, huge),
       big_cpy(buf, quo, huge),
       big_div(quo, divisor, huge))  /* Quotient relevant ? */	$ Y Y N
     (divisor & 2)			    /* +/- Zyklus ? */	$ Y N .
  $A big_sub(sum, quo, huge);	      /* von Summe abziehen */	$ X . .
     big_add(sum, quo, huge);	      /* zur Summe addieren */	$ . X .
     big_cpy(quo, buf, huge);			  /* merken */	$ X X .
     break;			     /* Iteration beenden ! */	$ . . X
  $END

  $TABLE DoCalcPh2		  /* Phase 2 der Berechnung */
  /*--------------------------------------------------------*/
     big_set(quo, 4*239, huge); /* init. Quotient auf 4*239 */
     for (divisor = 1; ; divisor += 2)	  /* Iter.schleife: */
  $C ( big_div(quo, 239*239, huge),
       big_cpy(buf,	quo, huge),
       big_div(quo, divisor, huge))  /* Quotient relevant ? */	$ Y Y N
     (divisor & 2)			    /* +/- Zyklus ? */	$ Y N .
  $A big_sub(sum, quo, huge);	      /* von Summe abziehen */	$ . X .
     big_add(sum, quo, huge);	      /* zur Summe addieren */	$ X . .
     big_cpy(quo, buf, huge);			  /* merken */	$ X X .
     break;			     /* Iteration beenden ! */	$ . . X
  $END

  $TABLE PrintNum		     /* Ende der Berechnung */
  /*--------------------------------------------------------*/
     big_out(sum, digits, huge);       /* Resultat ausgeben */
     goto DoActArg;	     /* zur naechsten Anforderung ! */
  $END
}

void big_set(field, value, length)
  register unsigned *field, value;
  register int length;
{ 
  $TABLE BigSet 		/* Initialisiere Rechenfeld */
  /*--------------------------------------------------------*/
     *(field++) = value;	    /* uebernehme Kopf-Wert */
     for (;;)			     /* Iterationsschleife: */
  $C (--length) 			   /* noch zu tun ? */	$ Y N
  $A *(field++) = 0;			   /* Rest loeschen */	$ X .
     break;			     /* Iteration beenden ! */	$ . X
  $END
}

void big_cpy(field, source, length)
  register unsigned *field, *source;
  register int length;
{ 
  $TABLE BigCopy		      /* Kopiere Rechenfeld */
  /*--------------------------------------------------------*/
      for (;;)			     /* Iterationsschleife: */
  $C (--length) 			   /* noch zu tun ? */	$ Y N
  $A *(field++) = *(source++);		/* jeweils kopieren */	$ X .
     break;			     /* Iteration beenden ! */	$ . X
  $END
}

int big_div(field, divisor, length)
  unsigned *field;
  register unsigned divisor;
  int length;
{ register unsigned long accumulator;
  register unsigned *p_work, *p_bound;
  register int any_bits;

  $TABLE BigDiv 		 /* Division auf Rechenfeld */
  /*--------------------------------------------------------*/
     accumulator = 0L;		 /* Zwischensumme noch leer */
     any_bits = 0;		 /* noch nichts verarbeitet */
     p_work = field;		    /* setze Arbeits-Zeiger */
     p_bound = &p_work[length];       /* setze Grenz-Zeiger */
     for (; ; ++p_work) 	     /* Iterationsschleife: */
  $C (p_work < p_bound) 		   /* noch zu tun ? */	$ Y Y Y N
     ((accumulator += *p_work) != 0L)	     /* Teilsumme ? */	$ Y Y N .
     ((*p_work = (unsigned)(accumulator/divisor)) != 0)
					  /* Teilresultat ? */	$ Y N . .
  $A any_bits = 1;		      /* merke Verarbeitung */	$ X . . .
     accumulator %= divisor;
     accumulator <<= (8*sizeof(unsigned));	/* schiften */	$ X X . .
     return any_bits;		      /* Ende und Antwort ! */	$ . . . X
  $END
}

void big_add(field, summand, length)
  unsigned *field, *summand;
  int length;
{ register unsigned long accumulator;
  register unsigned *p_fld, *p_smd;

  $TABLE BigAdd 		 /* Addition auf Rechenfeld */
  /*--------------------------------------------------------*/
     accumulator = 0L;		 /* Zwischensumme noch leer */
     p_fld = &field[length];
     p_smd = &summand[length];	    /* setze Arbeits-Zeiger */
     for (;;)			     /* Iterationsschleife: */
  $C ((--p_smd, --p_fld) >= field)	   /* noch zu tun ? */	$ Y N
  $A accumulator >>= (8*sizeof(unsigned));
     accumulator += *p_fld;
     accumulator += *p_smd;
     *p_fld = (unsigned)accumulator;	   /* Teil-Addition */	$ X .
     break;			     /* Iteration beenden ! */	$ . X
  $END
}

void big_sub(field, subtrahend, length)
  unsigned *field, *subtrahend;
  int length;
{ register unsigned long accumulator;
  register unsigned *p_fld, *p_std;

  $TABLE BigAdd 		 /* Addition auf Rechenfeld */
  /*--------------------------------------------------------*/
     accumulator = 0L;		 /* Zwischensumme noch leer */
     p_fld = &field[length];
     p_std = &subtrahend[length];   /* setze Arbeits-Zeiger */
     for (;;)			     /* Iterationsschleife: */
  $C ((--p_std, --p_fld) >= field)	   /* noch zu tun ? */	$ Y N
  $A accumulator >>= (8*sizeof(unsigned));
     accumulator += *p_fld;
     accumulator += ~*p_std;
     *p_fld = (unsigned)accumulator;	   /* Teil-Addition */	$ X .
     break;			     /* Iteration beenden ! */	$ . X
  $END
}

void big_mul(field, factor, length)
  unsigned *field, factor;
  int length;
{ register unsigned long accumulator, l_factor;
  register unsigned *p_work;

  $TABLE BigMult	   /* Multiplikation auf Rechenfeld */
  /*--------------------------------------------------------*/
     accumulator = 0L;		 /* Zwischensumme noch leer */
     l_factor = factor; 	      /* Faktor verlaengern */
     p_work = &field[length];	    /* setze Arbeits-Zeiger */
     for (;;)			     /* Iterationsschleife: */
  $C (--p_work >= field)		   /* noch zu tun ? */	$ Y N
  $A accumulator >>= (8*sizeof(unsigned));
     accumulator += *p_work * l_factor;
     *p_work = (unsigned)accumulator;/* Teil-Multiplikation */	$ X .
     break;			     /* Iteration beenden ! */	$ . X
  $END
}

void big_out(field, digits, length)
  register unsigned *field;
  register int digits, length;
{ register int count;

  $TABLE BigOut 		 /* Ausgabe des Rechenfelds */
  /*--------------------------------------------------------*/
     for (count = 0, digits += (3-1); ; count += 3)
				     /* Iterationsschleife: */
  $C (count < digits)			   /* noch zu tun ? */	$ Y N
  $A printf("%03u",field[0]);
     printf(!count ? "." : ((count+3) % (18*3)) ? " " : "\n");
     field[0] = 0;
     big_mul(field, 1000, length);	    /* Teil-Ausgabe */	$ X .
     printf("\n\n");		      /* Ausgabe Trennzeile */	$ . X
     break;			     /* Iteration beenden ! */	$ . X
  $END
}
