/* ET-Beispiel-Hauptprogramm: (C) 1990/Jan/06 - 1991/Okt/16,    */
/*                             R. Scharnagl, 4000 Duesseldorf 1 */

  /*============================================================*/
  /* PRIM.T - Demonstrations-Hauptprogramm zur Benutzung von    */
  /* Entscheidungs-Tabellen (ETs) in C-Programmen mittels des   */
  /* Precompilers DTT.                                          */
  /*------------------------------------------------------------*/
  /* Eine ET hat hier formal das folgende Erscheinungsbild,     */
  /* (wobei eine evtl. ELSE-Spalte nur aus (.) besteht):        */
  /*                                                            */
  /* Start-Makro:       $TABLE <name>                           */
  /* Bedingungsmakro:   $C <cond.>      $ <flags (Y|N|.)> ...   */
  /*                       <cond.>      $ <flags (Y|N|.)> ...   */
  /*                       ...                                  */
  /* Aktionsmakro:      $A <statm.>     $ <flags (X|.)>   ...   */
  /*                       <statm.>     $ <flags (x|.)>   ...   */
  /*                       ...                                  */
  /* Ende-Makro:        $END                                    */
  /*                                                            */
  /* Es sind auch modifizierte ET-Formen moeglich. So duerfen   */
  /* zwischen $TABLE und $C eine Reihe von Anweisungen stehen   */
  /* (oder auch Schleifenkoepfe), die als gemeinsame Liste      */
  /* vorgeschalteter Aktionen aufgefasst wird. Die $C- und      */
  /* $A-Teile koennen auch gaenzlich fehlen, es liegt dann      */
  /* eine reine Aktionstabelle vor. Im $C-Teil kann das "if"    */
  /* des Bedingungsteils auch entfallen, es wird dann spaeter   */
  /* ergaenzt. Eine logische Zeile kann sich ueber mehrere      */
  /* physische Zeilen erstrecken. Innerhalb von ETs sind auch   */
  /* Kommentare, falls nicht mehrzeilig, erlaubt.               */
  /*------------------------------------------------------------*/
  /* DTT kann mit diversen Optionen aufgerufen werden:          */
  /*   -T  :  Trace generating (Zaehlung beim Programmlauf)     */
  /*   -C  :  Copy tables (ETs werden als Kommentar wiederholt) */
  /*   -S  :  Silent (alle Meldungen werden unterdrueckt)       */
  /*   -M  :  Mark ELSE usage (protokolliere ELSE-Benutztung)   */
  /*   -N  :  No ANSI standard (Fuer C-Compiler alten Typs)     */
  /*------------------------------------------------------------*/
  /* Im Trace-Mode wird mit dem zugelinkten Modul TRACE.O nach  */
  /* dem Programm-Ende eine Uebersicht aller angesprochenen und */
  /* mit der Trace-Option uebersetzten Module aufgelistet.      */
  /* Die Anzahl der Programmdurchlaeufe jeder Tabellenregel ist */
  /* darin ersichtlich. Nicht vollstaendig durchlaufene ETs     */
  /* werden mit '?' markiert, unberuehrte durch '*'. So laesst  */
  /* sich die Unvollstendigkeit einer Testumgebung bzw. uner-   */
  /* reichbarer Programmcode leichter erkennen und abaendern.   */
  /* Auch beim Test von GEM-Anwendungen ist durch ein (DT_TRACE */
  /* bedingt compiliertes) Umlenken von 'stdout' vor dem Pro-   */
  /* grammende eine Dokumentation von Dialoganwendungssitzungen */
  /* in aehnlicher Weise moeglich.                              */
  /*------------------------------------------------------------*/
  /* DTT verifiziert die Korrektheit der ETs, die Vollstaendig- */
  /* keit der Regeln und gibt entsprechende Hinweise aus.       */
  /*============================================================*/

#ifdef __STDC__
#include <stdlib.h>
#define _(parms) parms /* prototyp */
#else
#define void  /* */
#define const /* */
#define _(parms) ()    /* old declaration */
#endif

/* Benoetigte Copy-Bibliothek */
#include <stdio.h>

#ifdef __TURBOC__
#pragma warn -use
#endif

/* Prototypen */
void    Ausgabe         _((int i));
void    Outc            _((char c));
void    Vorbereitung    _((void)); 
void    Hauptteil       _((void));
int     main            _((void));

/* Globale Variablen */
int     test, teil, quot, zaehler, anzahl;

int main()             /*=======================================*/
{                      /* Berechnung der ersten N.N. Primzahlen */
                       /*=======================================*/

  $TABLE Steuerung                     /* REINE AKTIONS-TABELLE */
  /*------------------------------------------------------------*/
     Vorbereitung();             /* Vorbereitungen durchfuehren */
     Hauptteil();             /* eigentliches Programm aufrufen */
     return 0;                                 /* normales Ende */
  $END
}

void Vorbereitung()    /*=======================================*/
{                      /* Einmalige Vorbereitungen durchfuehren */
                       /*=======================================*/

  $TABLE Titelzeile                    /* REINE AKTIONS-TABELLE */
  /*------------------------------------------------------------*/
     Ausgabe(anzahl = 260);                 /* Anzahl festlegen */
     puts(" PRIMZAHLEN:");         /* die Ueberschrift ausgeben */
  $END

  $TABLE Grundwert                           /* ET MIT 2 REGELN */
  /*------------------------------------------------------------*/
  $C (test == 0)       /* eigentlich ueberfluessig, immer: true */  $ Y N
  $A test = 1;                      /* setze die erste Testzahl */  $ X .
     zaehler = 0;             /* setze Primzahl-Zaehler zurueck */  $ X .
  $END
}

void Hauptteil()                /*==============================*/
{                               /* Eigentlicher Primzahl-Zyklus */
                                /*==============================*/

  $TABLE neue_Zahl                           /* ET MIT 2 REGELN */
  /*------------------------------------------------------------*/
  $C (test > 2)           /* gerade Primzahl schon unmoeglich ? */  $ N Y
  $A test += 1;                      /* dann zur naechsten Zahl */  $ X .
     test += 2;            /* hier zur naechsten ungeraden Zahl */  $ . X
     teil = 1;       /* setze ersten (verbrauchten) Teiler fest */  $ X X
  $END

  $TABLE neuer_Test          /* ET MIT VORAKTIONEN UND 3 REGELN */
  /*------------------------------------------------------------*/
     teil += 1;               /* Uebergang zum naechsten Teiler */
     quot = test / teil;             /* berechne den Quotienten */
  $C (quot < teil)             /* Wurzel schon ueberschritten ? */  $ Y N N
     (quot * teil == test)                /* war Zahl teilbar ? */  $ . Y N
  $A goto $T;           /* $T-Makro: wiederhole die ET von vorn */  $ . . X
     goto neue_Zahl;            /* neuen Test-Kandidaten suchen */  $ . X .
  $END

  $TABLE Primzahl   /* ET MIT VORAKTIONEN UND 3 REGELN MIT ELSE */
  /*------------------------------------------------------------*/
     zaehler += 1;                      /* zaehle neue Primzahl */
     Ausgabe(test);                  /* drucke die Primzahl aus */
  $C if (zaehler < anzahl)      /* noch nicht genug ermittelt ? */  $ N Y .
     if (zaehler % 13)            /* hat die Zeile noch Platz ? */  $ . Y .
  $A Outc('\n');                           /* die Zeile beenden */  $ X . X
     goto neue_Zahl;              /* suche naechsten Kandidaten */  $ . X X
     return;                                          /* fertig */  $ X . .
  $END

  $TABLE Unsinn               /* UEBERFLUESSIGE AKTIONS-TABELLE */
  /*------------------------------------------------------------*/
     return;                    /* kann niemals erreicht werden */
  $END
}

